import json
import os
import time
import portalocker
from datetime import datetime, timedelta

class FileTokenManager:
    def __init__(self, filepath: str):
        self.filepath = filepath
        os.makedirs(os.path.dirname(filepath) or ".", exist_ok=True)

    def _read_token(self):
        if not os.path.exists(self.filepath):
            return None
        with portalocker.Lock(self.filepath, 'r', timeout=2) as f:
            return json.load(f)

    def _write_token(self, token_data):
        with portalocker.Lock(self.filepath, 'w', timeout=2) as f:
            json.dump(token_data, f)

    def set_token(self, access_token, refresh_token, expires_in: int = 3600):
        expiry_time = (datetime.utcnow() + timedelta(seconds=expires_in)).isoformat()
        token_data = {
            "access_token": access_token,
            "refresh_token": refresh_token,
            "expires_at": expiry_time
        }
        self._write_token(token_data)

    def get_token(self):
        token_data = self._read_token()
        return token_data.get("access_token") if token_data else None

    def is_token_valid(self):
        token_data = self._read_token()
        if not token_data or "expires_at" not in token_data:
            return False
        return datetime.fromisoformat(token_data["expires_at"]) > datetime.utcnow()

    def clear_token(self):
        if os.path.exists(self.filepath):
            try:
                os.remove(self.filepath)
            except Exception as e:
                print(f"Warning: Could not delete token file: {e}")

    def refresh_token_with_lock(self, refresh_func):
        """Centralized refresh logic. Only one process will refresh; others will wait."""
        lockfile = self.filepath + ".lock"
        with portalocker.Lock(lockfile, timeout=60):
            # Check again inside lock in case another process already refreshed
            if self.is_token_valid():
                return
            # Refresh token using the passed-in callback (ABRestProcessor)
            access_token, refresh_token, expires_in = refresh_func()
            self.set_token(access_token, refresh_token, expires_in)

    def ensure_token_ready(self, refresh_callback=None):
        """Ensure that a valid token is available. Only one process refreshes it."""
        if self.is_token_valid():
            return

        lock_file = self.filepath + ".lock"

        # Lock ensures only one process can refresh
        with portalocker.Lock(lock_file, timeout=30):
            if not self.is_token_valid():
                print("[TokenManager] Token is invalid. Refreshing...")
                if refresh_callback:
                    refresh_callback()
            else:
                print("[TokenManager] Another process already refreshed the token.")




class PurchaseInvoiceApiUrl:
    PURCHASE_INVOICE = "/api/v3/purchase-invoices"
    PURCHASE_INVOICE_DETAILS = "/api/v3/purchase-invoices/{id}"
    PURCHASE_INVOICE_LINE_DETAILS = "/api/v3/purchase-invoices/{id}/lines/{uuid}"
    PURCHASE_INVOICE_ACCOUNT_DETAILS = "/api/v3/accounts/{id}/purchase-invoices"
    PURCHASE_INVOICE_CREATE = "/api/v3/purchase-invoices/"
    PURCHASE_INVOICE_CANCEL = "/api/v3/purchase-invoices/{id}/cancel"
    PURCHASE_INVOICE_REACTIVE = "/api/v3/purchase-invoices/{id}/reactive"
    PURCHASE_INVOICE_DELETE = "/api/v3/purchase-invoices/{id}"
    PURCHASE_INVOICE_AMEND = "/api/v3/purchase-invoices/{purchase_invoice_id}/amend"
    PURCHASE_INVOICE_INFORMATION = "/api/v3/purchase-invoices/{purchase_invoice_id}/information"
    PURCHASE_INVOICE_FROM_PURCHASE_ORDER = "/api/v3/purchase-orders/{purchase_order_id}/purchase-invoices"



from dataclasses import dataclass
from typing import List
from exsited.exsited.account.dto.account_nested_dto import *
from exsited.exsited.common.dto.common_dto import *
from exsited.exsited.order.dto.order_dto import OrderDataDTO
from exsited.exsited.order.dto.order_nested_dto import OrderPropertiesDTO
from exsited.sdlize.ab_base_dto import ABBaseDTO


@dataclass(kw_only=True)
class PricingRuleDTO(ABBaseDTO):
    price: str = None


@dataclass(kw_only=True)
class ItemPriceSnapshotDTO(ABBaseDTO):
    pricingRule: PricingRuleDTO = None


@dataclass(kw_only=True)
class ItemPriceTaxDTO(ABBaseDTO):
    uuid: str = None
    code: str = None
    rate: float = None


@dataclass(kw_only=True)
class PaymentAppliedDTO(ABBaseDTO):
    processor: str = None
    amount: str = None


@dataclass(kw_only=True)
class PaymentDTO(ABBaseDTO):
    paymentApplied: List[PaymentAppliedDTO] = None


@dataclass(kw_only=True)
class InvoiceDTO(ABBaseDTO):
    payment: PaymentDTO = None


@dataclass(kw_only=True)
class ContractPropertiesDTO(ABBaseDTO):
    requireCustomerAcceptance: str = None
    requiresPaymentMethod: str = None
    initialContractTerm: str = None
    renewAutomatically: str = None
    autoRenewalTerm: str = None
    allowEarlyTermination: str = None
    applyEarlyTerminationCharge: str = None
    allowPostponement: str = None
    maximumDurationPerPostponement: str = None
    maximumPostponementCount: str = None
    allowTrial: str = None
    startContractAfterTrialEnds: str = None
    trialPeriod: str = None
    allowDowngrade: str = None
    periodBeforeDowngrade: str = None
    allowDowngradeCharge: str = None
    downgradeChargeType: str = None
    downgradeChargeFixed: str = None
    allowUpgrade: str = None


@dataclass(kw_only=True)
class OrderLineDTO(ABBaseDTO):
    itemId: str = None
    itemOrderQuantity: int = None
    itemPriceSnapshot: ItemPriceSnapshotDTO = None
    itemPriceTax: ItemPriceTaxDTO = None
    packageName: str = None


@dataclass(kw_only=True)
class OrderDTO(ABBaseDTO):
    lines: List[OrderLineDTO] = None
    invoice: InvoiceDTO = None
    allowContract: str = None
    contractProperties: ContractPropertiesDTO = None
    properties: OrderPropertiesDTO = None


@dataclass(kw_only=True)
class AccountDTO(ABBaseDTO):
    name: str = None
    emailAddress: str = None
    status: str = None
    id: str = None
    displayName: str = None
    description: str = None
    invoiceParentAccount: str = None
    type: str = None
    imageUri: str = None
    grantPortalAccess: str = None
    website: str = None
    linkedin: str = None
    twitter: str = None
    facebook: str = None
    createdBy: str = None
    createdOn: str = None
    lastUpdatedBy: str = None
    lastUpdatedOn: str = None
    uuid: str = None
    version: str = None
    parentAccount: str = None
    group: str = None
    manager: str = None
    referralTracking: str = None
    salesRep: str = None
    pricingLevel: PricingLevelDTO = None

    currency: CurrencyDTO = None
    timeZone: TimeZoneDTO = None
    tax: TaxDTO = None
    accountingCode: AccountingCodeDTO = None
    communicationPreference: list[CommunicationPreferenceDTO] = None
    paymentMethods: list[PaymentMethodsDataDTO] = None
    billingPreferences: BillingPreferencesDTO = None
    customAttributes: list[CustomAttributesDTO] = None
    try:
        addresses: list[AcccountAddressDTO] = None
        contacts: list[ContactDTO] = None
    except:
        addresses: list[AcccountAddressRequestDTO] = None
        contacts: list[ContactRequestDTO] = None
    customForms: CustomFormsDTO = None
    eventUuid: str = None
    customObjects: list[CustomObjectDTO] = None
    kpis: dict = None
    id: str = None
    try:
        order: OrderDTO = None
    except:
        order: OrderDataDTO = None


@dataclass(kw_only=True)
class ExpressDTO(ABBaseDTO):
    account: AccountDTO = None
    _custom_field_mapping = {
        "isTaxExemptWhenSold": "isTaxExemptWhenSold"
    }